package QIF;

use Fcntl; # For O_RDWR, O_CREAT, etc.
use SDBM_File;
use PayPal::DB;
use JSON;
use Carp;
use Term::ReadKey;

# create string for QIF account record
# @param $type account type string.
# @param $name account name string.
# @param $desc account description string.
# @return a string.
sub account
{
  my ($type, $name, $desc) = @_;
  return '' unless $type;
  return '' unless $name;
  my $a = "!Account\n";
  $a .= "N$name\n";
  $a .= "T$type\n";
  $a .= "D$desc\n" if $desc;
  $a .= "^\n";
  return $a;
}

# create transaction type ID.
# @param $txn reference to transaction hash.
# @return string with transaction type ID.
sub txn_type_id
{
  my ($txn) = @_;
  return unless ref($txn) eq 'HASH';
  my $a = "$txn->{Type}/$txn->{TransactionType}/";
  if (! ref($txn->{PII_Name})) {
    $a .= $txn->{PII_Name};
  } elsif (ref($txn->{PII_Name}) eq 'SCALAR') {
    $a .= ${$txn->{PII_Name}};
  } else {
    $a .= ref($txn->{PII_Name});
  }
  return $a;
}

sub PII_Name
{
  my ($h) = @_;
  return '' unless $h;
  return $h unless ref($h);
  return $$h if ref($h) eq 'SCALAR';
  return join(', ', @$h) if ref($h) eq 'ARRAY';
  return "ref=".ref($h);
}

my $paypal_type2account_name_fn = $PayPal::DB::paypal_db_dir->() . 'paypal_type2account_name';
our %paypal_type2account_name;
tie(%paypal_type2account_name, 'SDBM_File', $paypal_type2account_name_fn, O_RDWR|O_CREAT, 0666) or die "Couldn't tie SDBM file '$paypal_type2account_name_fn': $!";
END { untie %paypal_type2account_name }

sub dump_paypal_type2account_name
{
    while(my ($k,$v) = each(%paypal_type2account_name)) {
	print "$k\t$v\n";
    }
}

# look up account name for a transaction
# @param $txn reference to transaction hash.
# @return string with transaction type ID, undef if no account name was found.
sub lookup_account_name
{
  my ($txn) = @_;
  return unless ref($txn) eq 'HASH';
  my $k = txn_type_id($txn);
  return unless $k;
  return unless exists($paypal_type2account_name{$k});
  return $paypal_type2account_name{$k};
}

use constant SHOW_SUMMARY => -1;
use constant SHOW_DETAILS => -2;
use constant SHOW_FULL_DETAILS => -3;

sub choose_account_name
{
    my @a = @_;

    # print @a
    for(my $i = 0; $i <= $#a; ++$i) {
	my $is = $i + 1;
	print "${is}: $a[$i]\n";
    }
    local $| = 1; # auto flush STDOUT
    print "\n> (number, new name, 'd' details, 'f' full details, 's' summary): ";

    # select an account
    my $i = ReadLine(0);
    chomp $i;
    unless (length($i)) {
	print STDERR "no account choosen\n";
	return
    }

    if ($i eq 's') {
	return SHOW_SUMMARY;
    }
    if ($i eq 'd') {
	return SHOW_DETAILS;
    }
    if ($i eq 'f') {
	return SHOW_FULL_DETAILS;
    }

    # if only digits were entered, lookup account from list.
    if ($i =~ /^\d+$/) {
	$i += 0; # make integer out of string
	if ($i == 0) {
	    print STDERR "zero not valid";
	    return
	}
	--$i; # make zero based index
	if ($i > $#a) {
	    print STDERR "invalid number";
	    return
	}
	return $a[$i];
    }

    # use input as account name directly
    if ($i =~ /^\s+$/) {
	print STDERR "empty account name";
	return
    }
    return $i;
}

# sort the parameter list into unique values.
# @param @_ list of (scalar) values.
# @return sorted @_ with duplicates removed.
sub sort_unique
{
    my %a;
    foreach (@_) {
	$a{$_} = 1;
    }
    return keys %a;
}

# display transaction details and let user choose an account name.
# newly entered account names are saved in a persistent state file.
# @param $txn reference to transaction hash.
# @param @account_names optional additional account names.
sub interactiveAccountSelect
{
    my $txn = shift;
    croak('$txn parameter is not a hash reference') unless ref($txn) eq 'HASH';
    my @account_names = sort_unique(values(%paypal_type2account_name), @_);

    my $res = SHOW_SUMMARY;
    while ($res < 0) {
	print "\n\nSelect a Quickbooks account name for the following transaction:\n";
	if ($res == SHOW_SUMMARY) {
	    my $pii_name = PII_Name($txn->{PII_Name});
	    print<<"EOT";
Type:  $txn->{Type} $txn->{TransactionType} $pii_name
Gross: $txn->{GrossAmount}
Name:  $txn->{PayerDisplayName} $txn->{Payer}
EOT
	} elsif ($res == SHOW_DETAILS) {
	    foreach my $k (sort keys %$txn) {
		# only print if key and value are not empty
		next unless length($k) > 0;
		my $v = $txn->{$k};
		next unless length($v) > 0;
		print "${k}:\t$v\n";
	    }
	} elsif ($res == SHOW_FULL_DETAILS) {
	    print to_json($txn, {utf8 => 1, pretty => 1});
	} else {
	    croak("invalid \$res value: $res");
	}
	$res = choose_account_name(@account_names);
    }

    # save selected account name into paypal_type2account_name hash for persistence
    my $k = txn_type_id($txn);
    $paypal_type2account_name{$k} = $res;

    return $res;
}

sub dump_txn
{
  my ($txn) = @_;
  die "dump_txn parameter is not a hash reference" unless ref($txn) eq 'HASH';
  print STDERR to_json($txn, {utf8 => 1, pretty => 1}) . "\n";
}

sub enforce_sign
{
  my ($a) = @_;
  return '+' . $a if $a >= 0.0;
  return $a;
}

# create a QIF transaction string.
# \todo see how taxes should be handled.
# @param txn transaction hash reference.
# @param $account_name account name string of bank account.
# @param $fee_name Expense account name of fee account.
# @return a string.
sub transaction
{
  my ($txn, $account_name, $fee_name) = @_;
  return unless ref($txn) eq 'HASH';
  return unless $account_name;
  return unless $fee_name;

  unless ($h->{NetAmount} == $h->{GrossAmount} + $h->{FeeAmount}) {
    dump_txn($txn);
    croak("NetAmount==GrossAmount+FeeAmount does not match: $h->{NetAmount} "."=="." $h->{GrossAmount} "."+"." $h->{FeeAmount}");
  }

  my $grossamount = enforce_sign($txn->{GrossAmount});
  my $netamount = enforce_sign($txn->{NetAmount});
  my $feeamount = enforce_sign($txn->{FeeAmount});

  my $ts = $txn->{Timestamp};
  $ts ||= $txn->{PaymentDate};
  unless ($ts) {
    dump_txn($txn);
    croak("could not determine transaction time stamp");
  }

  my $a = "D$ts\n";

  $a .= "T$netamount\n";

  $a .= "N$txn->{TransactionID}\n" if $txn->{TransactionID};

  my $memo = "$txn->{Type} $txn->{TransactionType} $txn->{PayerDisplayName} " . PII_Name($txn->{PII_Name});
  $a .= "M$memo\n";

  $a .= "P$txn->{PayerDisplayName} $txn->{Payer}\n";
  #AAddress 1
  #AAddress 2

  $a .= "A$account_name\n";

  unless ($txn->{GnuCashAccountName}) {
    dump_txn($txn);
    croak("transaction does not contain GnuCashAccountName key");
  }
  $a .= "S$txn->{GnuCashAccountName}\n";
  $a .= "E$memo\n";
  $a .= "\$$grossamount\n";

  if ($txn->{FeeAmount} != 0.0) {
    $a .= "S$fee_name\n";
    $a .= "E$memo\n";
    $a .= "\$$feeamount\n";
  }

  $a .= "^\n";
  return $a;
}

##############################################################################
# command line

# strings for command line arguments
my $arg__dump_paypal_type_to_accountname = '--dump-paypal-type-to-accountname';
my $arg__clear_paypal_type_to_accountname = '--clear-paypal-type-to-accountname';

sub command_line_help
{
    my ($prog) = @_;
    print<<"EOT";

usage: $prog $arg__dump_paypal_type_to_accountname
  Print the hash which maps PayPal transaction types to Account names.

usage: $prog $arg__clear_paypal_type_to_accountname
  Remove all values from the hash which maps PayPal transaction types to
  Account names.
EOT
    return 1;
}

sub parse_command_line
{
    foreach my $arg (@_) {
	if ($arg eq $arg__dump_paypal_type_to_accountname) {
	    dump_paypal_type2account_name();
	    exit 0;
	}
	if ($arg eq $arg__clear_paypal_type_to_accountname) {
	    %paypal_type2account_name = ();
	    exit 0;
	}
    }
}

1;
