#!/usr/bin/env perl
# @(#) A program to download PayPal transactions and save them as IIF files.
# $Id: paypal2iif.pl 7924 2019-01-20 21:59:02Z doj $
#
# Copyright (c) 2013 Dirk Jagdmann <doj@cubic.org>
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
#    1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgement in the product documentation would be
#    appreciated but is not required.
#
#    2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
#
#    3. This notice may not be removed or altered from any source
#    distribution.
#
# For more details visit the website https://llg.cubic.org/csv2iif

use lib './';
use configuration;
use JSON;
use Business::PayPal::API qw( GetTransactionDetails TransactionSearch );
use PayPal;
use PayPal::DB;
use Quickbooks::IIF;
use POSIX;
use Time::Local;

my $iif;
my $do_help = 0;

my $end_ts = time();
my $start_ts = $end_ts - 86400*10;
if (PayPal::DB::get_state('last_paypal_API_sync') > 0)
{
    $start_ts = PayPal::DB::get_state('last_paypal_API_sync') - 1;
    die "strange start ts $start_ts" unless $start_ts > 100;
}

# parse command line
PayPal::DB::parse_command_line(@ARGV);
foreach my $arg (@ARGV) {
    if ($iif eq 'fetch me') {
	$iif = $arg;
	next
    }
    if ($arg eq '-o') {
	$iif = 'fetch me';
	next
    }

    if ($start_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --start, needs to be 'YYYY-MM-DD'";
	}
	$start_ts = timelocal(0, 0, 0, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --start" unless $start_ts > 100;
    }
    if ($arg eq '--start') {
	$start_ts = 'fetch me';
	next;
    }

    if ($end_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --end, needs to be 'YYYY-MM-DD'";
	}
	$end_ts = timelocal(59, 59, 23, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --end" unless $end_ts > 100;
    }
    if ($arg eq '--end') {
	$end_ts = 'fetch me';
	next;
    }

    if ($arg eq '--lastsync') {
	my $ts = PayPal::DB::get_state('last_paypal_API_sync');
	if ($ts > 0) {
	    print POSIX::ctime($ts);
	} else {
	    print "not synced with PayPal yet.\n";
	}
	exit 0;
    }

    if ($arg eq '--help' || $arg eq '-h') {
	$do_help = 1;
	next;
    }
}

help() if $start_ts eq 'fetch me';
help() if $end_ts eq 'fetch me';
die "start day need to be before end day" if $start_ts > $end_ts;

if (! defined($iif)) {
    my ($s_sec,$s_min,$s_hour,$s_mday,$s_mon,$s_year) = localtime($start_ts);
    my ($e_sec,$e_min,$e_hour,$e_mday,$e_mon,$e_year) = localtime($end_ts);
    $iif = sprintf("PayPal_%04i-%02i-%02iT%02i%02i%02i_%04i-%02i-%02iT%02i%02i%02i.iif",
		   $s_year+1900, $s_mon+1, $s_mday, $s_hour, $s_min, $s_sec,
		   $e_year+1900, $e_mon+1, $e_mday, $e_hour, $e_min, $e_sec,
		  );
}

help() if $do_help;

my $pp = new Business::PayPal::API(
				   Username   => $configuration::paypal_username,
				   Password   => $configuration::paypal_password,
				   Signature  => $configuration::paypal_signature,
				   sandbox    => $configuration::paypal_sandbox
				  ) or die "could not init PayPal API";

my $start_str = PayPal::ts2str($start_ts);
my $end_str = PayPal::ts2str($end_ts);
print "getting PayPal transactions from $start_str to $end_str ...\n";

$| = 1; # auto flush STDOUT
my @txn;
my $total = 0;
while(1)
{
    $end_str = PayPal::ts2str($end_ts);
    #print "# requesting $start_str to $end_str\n\n";
    my $a = $pp->TransactionSearch( StartDate => $start_str, EndDate => $end_str );

    my $z = 0;
    my $last_ts = PayPal::str2ts('2038-01-01T00:00:00Z'); # \todo fix sometime in 2037
    foreach my $h (@$a) {
	++$z;
	++$total;
	my $txn = $h->{TransactionID};
	if (PayPal::DB::get_state($txn)) {
	    print STDERR "skipping $txn from $h->{Timestamp} converted ".asctime(localtime(PayPal::DB::get_state($txn)));
	    next;
	}

	my $ts = PayPal::str2ts($h->{Timestamp});
	die "time stamp is not less than previous time stamp" unless $ts <= $last_ts;
	$last_ts = $ts;

	my %t;
	my $t = PayPal::DB::get_txn($txn);
	if (ref($t) eq 'HASH') {
	    %t = %$t;
	} else {
	    %t = $pp->GetTransactionDetails( TransactionID => $txn );
	}
	# merge $h into %t
	@t{keys %$h} = values %$h;

	#print STDERR "##########################################################\n";
	#print STDERR "# $txn\n";
	#print STDERR to_json(\%t, {utf8 => 1, pretty => 1}) . "\n";

	# rename keys
	if ($t{Type} eq "Refund") {
	    # do nothing
	} elsif ($t{Type} eq "Transfer") {
	    $t{Title} = 'Transfer to Back Account';
	    $t{Type} = "Withdraw Funds to a Bank Account";
	} elsif ($t{Type} eq "Fee Reversal") {
	    undef $txn; # don't convert this to IIF
	} elsif ($t{TransactionType} eq "cart") {
	    $t{Title} = 'Shopping Cart';
	} elsif ($t{TransactionType} eq "web-accept") {
	    if (ref($t{PII_Name})) {
		die "$txn unexpected PII_Name type for TransactionType web-accept";
	    }
	    $t{Title} = $t{PII_Name};
	} elsif ($t{TransactionType} eq "send-money") {
	    $t{Title} = $t{TransactionType};
	} else {
	    print STDERR to_json(\%t, {utf8 => 1, pretty => 1}) . "\n";
	    die "$txn: unknown Type: '$t{Type}' and TransactionType: '$t{TransactionType}'";
	}

	push(@txn, $txn) if $txn;
	PayPal::DB::save_txn(\%t) or die "$txn could not save PayPal transaction";

	print "\r$total ";
    }

    last if $z < 100;
    die "more than 100 transaction on ".PayPal::ts2str($last_ts) if $end_ts == $last_ts;
    $end_ts = $last_ts + 1; # since there may be multiple transactions on any time stamp, move end_ts out 1 second to catch them all
}
print "\rfetched $total transactions from PayPal.\n";

my $res = PayPal::iif($iif, @txn);
PayPal::DB::set_state('last_paypal_API_sync', $end_ts) or die "could not set sync time state";
if ($res == 0) {
    print "No transactions saved.\n";
    exit 0;
}
print "saved $res transactions to $iif\n";

PayPal::DB::set_states($end_ts, @txn) or die "could not set txn state";
exit 0;

sub help
{
    my $prog = 'paypal2iif.pl';
    print<<"EOT";
usage: $prog
  Fetch transactions from PayPal and convert to $iif

usage: $prog -o out.iif
  Fetch transactions from PayPal and convert to out.iif

usage: $prog --start 'YYYY-MM-DD'
  Fetch transactions from PayPal starting at midnight from the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --end 'YYYY-MM-DD'
  Fetch transactions from PayPal until 23:59:59 at the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --lastsync
  Show timestamp of last sync with PayPal.
EOT
    PayPal::DB::command_line_help($prog);
    exit 1;
}
